# -*- coding: utf-8 -*-

import json


def _escape_for_embedding(json_string):
    return json_string.replace('\\', '\\\\').replace("'", "\\'")

def escape_json_or_null(json_string):
    """Escape the JSON string. If None, then return JSON 'null'."""
    if json_string is None:
        return json.dumps(None)
    else:
        return _escape_for_embedding(json_string)

def json_dump_and_escape(data):
    return _escape_for_embedding(json.dumps(data))



from __future__ import print_function, unicode_literals

from escher.plots import Builder
from escher.urls import get_url
from escher.version import __version__
from escher.urls import top_directory, root_directory

from os.path import join, dirname, realpath
from jinja2 import Environment, PackageLoader
import shutil

# set up jinja2 template location
env = Environment(loader=PackageLoader('escher', 'templates'))

def generate_static_site():
    print('Generating static site at %s' % top_directory)

    # index file
    template = env.get_template('homepage.html')

    def static_rel(path):
        return 'py/' + path

    data = template.render(d3=static_rel(get_url('d3', 'local')),
                           boot_css=static_rel(get_url('boot_css', 'local')),
                           homepage_css=static_rel(get_url('homepage_css', 'local')),
                           favicon=static_rel(get_url('favicon', 'local')),
                           logo=static_rel(get_url('logo', 'local')),
                           documentation=get_url('documentation', protocol='https'),
                           github=get_url('github', protocol='https'),
                           github_releases=get_url('github_releases', protocol='https'),
                           homepage_js=static_rel(get_url('homepage_js', 'local')),
                           version=__version__,
                           map_download_url=get_url('map_download', 'local'),
                           web_version=True,
                           server_index_url=static_rel(get_url('server_index', 'local')))

    with open(join(top_directory, 'index.html'), 'wb') as f:
        f.write(data.encode('utf-8'))

    # viewer and builder
    # make the builder
    builder = Builder(safe=True, id='static_map')

    filepath = join(top_directory, 'builder')
    with open(join(root_directory, get_url('server_index', source='local')), 'r') as f:
        index_json = f.read()
    html = builder.save_html(filepath=filepath,
                             overwrite=True,
                             js_source='local',
                             protocol=None,
                             minified_js=True,
                             static_site_index_json=index_json)

    # copy over the source maps
    escher_map = get_url('escher_min', 'local') + '.map'
    builder_css_map = get_url('builder_css_min', 'local') + '.map'
    shutil.copy(join(root_directory, escher_map), join(top_directory, 'builder', escher_map))
    shutil.copy(join(root_directory, builder_css_map), join(top_directory, 'builder', builder_css_map))


if __name__=='__main__':
    generate_static_site()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This script has two purposes:

1. Apply the attributes from a cobra model to an existing Escher
map. For instance, update modified reversibilities.

2. Convert maps made with Escher beta versions to valid jsonschema/1-0-0
maps.

"""


from __future__ import print_function, unicode_literals


usage_string = """
Usage:

./convert_map.py {map path} {model path}

OR

python -m escher.convert_map {map path} {model path}
"""


try:
    import cobra.io
    import jsonschema
except ImportError:
    raise Exception(('The Python packages jsonschema and COBRApy (0.3.0b3 or later) '
                     'are required for converting maps.'))
import sys
import json
import random
import string
import hashlib
from os.path import basename, join
import logging
try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from escher.validate import validate_map, genes_for_gene_reaction_rule

# configure logger
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


def main():
    """Main entrypoint for convert_map. Instructions are at the top of this file."""
    # get the arguments
    try:
        in_file = sys.argv[1]
        model_path = sys.argv[2]
    except IndexError:
        print(usage_string)
        sys.exit()

    # load the cobra model
    try:
        model = cobra.io.load_json_model(model_path)
    except (IOError, ValueError):
        try:
            model = cobra.io.read_sbml_model(model_path)
        except IOError:
            raise Exception('Could not load the model: %s' % model_path)

    # get the current map
    with open(in_file, 'r') as f:
        out = json.load(f)

    # convert the map
    the_map = convert(out, model)

    # don't replace the file
    out_file = in_file.replace('.json', '_converted.json')
    logging.info('Saving validated map to %s' % out_file)
    with open(out_file, 'w') as f:
        json.dump(the_map, f, allow_nan=False)

# ------------------------------------------------------------------------------
# Functions for manipulating Escher maps as Python objects
# ------------------------------------------------------------------------------

class MissingDefaultAttribute(Exception):
    pass

def make_map(header, body):
    return [header, body]

def get_header(a_map):
    return a_map[0]

def set_header(a_map, value):
    a_map[0] = value

def get_body(a_map):
    return a_map[1]

def set_body(a_map, value):
    a_map[1] = value

def get_nodes(body):
    return body['nodes']

def set_nodes(body, value):
    body['nodes'] = value

def get_reactions(body):
    return body['reactions']

def set_reactions(body, value):
    body['reactions'] = value

def is_valid_header(val):
    """Header must have these values."""
    return (isinstance(val, dict) and
            all(x in val for x in
                ['schema', 'homepage', 'map_name', 'map_id', 'map_description']))

def is_valid_body(val):
    """Body must be a dictionary."""
    return isinstance(val, dict)

def has_header_and_body(val):
    """Check for header and body."""
    return (isinstance(val, list) and len(val) == 2 and
            is_valid_header(get_header(val)) and is_valid_body(get_body(val)))

# ------------------------------------------------------------------------------
# Functions for fixing nested dictionaries & lists
# ------------------------------------------------------------------------------

def dict_with_required_elements(the_dict, required_attributes, get_default=None,
                                nullable=[], cast={}):
    """Remove unsupported elements and provide defaults for missing
    elements or elements with zero length (e.g. {}, []).

    Arguments
    ---------

    the_dict: A dictionary.

    required_attributes: The attributes required in the dictionary.

    get_default: A function that takes the attribute name and the current
    object, and returns a default value. If not function is provided, then
    MissingDefaultAttribute is raised when an attribute is not present.

    nullable: A list of attributes that can be None.

    cast: A dictionary of attributes for keys and functions for values.

    """
    if type(the_dict) is not dict:
        raise MissingDefaultAttribute('(Bad object)')

    def has_zero_length(o):
        """Returns True if o has a length and it is zero."""
        try:
            return len(o) == 0
        except TypeError:
            return False

    def current_otherwise_default(name):
        """Take the value in the current dict, or else provide a default."""
        if name not in the_dict or has_zero_length(the_dict[name]):
            if get_default is not None:
                default = get_default(name, the_dict)
                the_dict[name] = default
            else:
                raise MissingDefaultAttribute(name)
        elif the_dict[name] is None and name not in nullable:
            raise MissingDefaultAttribute('%s (is None)' % name)
        # casting
        try:
            new = cast[name](the_dict[name])
        except KeyError:
            pass
        else:
            the_dict[name] = new

    # map over the dict
    not_required = set(the_dict.keys())
    for name in required_attributes:
        current_otherwise_default(name)
        # remember the keys that are not required
        try:
            not_required.remove(name)
        except KeyError:
            pass

    # remove not required
    for key in not_required:
        del the_dict[key]

def map_over_dict_with_deletions(the_dict, fix_function):
    """Use this to map over dictionary. The fix function updates values, and returns
    None to delete a value.

    Returns the dictionary.

    Arguments
    ---------

    the_dict: A dictionary to fix.

    fix_function: A function that takes the dictionary key and value as
    arguments, and that returns None if the key is deleted, otherwise
    returns the updated value.

    """

    # this will update the dictionary and return a set of deleted ids
    deleted_keys = set()
    for key, val in list(the_dict.items()):
        updated_val = fix_function(key, val)
        if updated_val is None:
            # add to aggregation set
            deleted_keys.add(key)
            # delete the value
            del the_dict[key]
        else:
            # set the new value
            the_dict[key] = updated_val

    return the_dict

def list_of_dicts_with_required_elements(a_list, required_attributes,
                                         get_default=None, nullable=[],
                                         cast={}):
    """For a list (dictionary) of dictionaries with required attributes, check
    each one. Returns the new list.

    Arguments
    ---------

    a_list: A dictionary with values that are dictionaries.

    required_attributes: A list of required attributes for the internal
    dictionaries.

    get_default: A function that takes the attribute name and the current
    object, and returns a default value. If not function is provided, then
    MissingDefaultAttribute is raised when an attribute is not present.

    """

    def fix_a_value(val):
        try:
            dict_with_required_elements(val, required_attributes,
                                        nullable=nullable, cast=cast)
            return val
        except MissingDefaultAttribute:
            return None

    # fix each element in the list
    return [y for y in (fix_a_value(x) for x in a_list)
            if y is not None]

def collection_of_dicts_with_required_elements(collection, required_attributes,
                                               get_default=None, nullable=[],
                                               cast={}):
    """For a collection (dictionary) of dictionaries with required
    attributes, check each one. Returns the new collection.

    Arguments
    ---------

    collection: A dictionary with values that are dictionaries.

    required_attributes: A list of required attributes for the internal
    dictionaries.

    get_default: A function that takes the attribute name and the current
    object, and returns a default value. If not function is provided, then
    MissingDefaultAttribute is raised when an attribute is not present.

    """

    def fix_a_value(val):
        try:
            dict_with_required_elements(val, required_attributes,
                                         nullable=nullable, cast=cast)
            return val
        except MissingDefaultAttribute:
            return None

    # update the dictionary
    for k, v in list(collection.items()):
        new_value = fix_a_value(v)
        if new_value is None:
            del collection[k]
        else:
            collection[k] = new_value

    return collection

# ------------------------------------------------------------------------------
# Functions for cleaning up unconnected map elements
# ------------------------------------------------------------------------------

def remove_unconnected_nodes(nodes, reactions, node_ids_deleted=set()):
    """Check for nodes with no connected segments in the reactions object, and
    return a new nodes object with only connected nodes.

    Arguments
    ---------

    nodes: A collection (dict) of nodes.

    reactions: A collection (dict) of reactions.

    node_ids_deleted: A set of previously deleted ids to update.

    """
    # update
    def map_over_connected_node_ids(fn, reactions):
        """Run fn with all node ids that have connected segments."""
        for reaction in reactions.values():
            for segment in reaction['segments'].values():
                fn(segment['from_node_id'])
                fn(segment['to_node_id'])

    # use a set to keep track of connected nodes
    nodes_with_segments = set()
    add_node = nodes_with_segments.add
    has_node = lambda x: x in nodes_with_segments
    # update
    map_over_connected_node_ids(add_node, reactions)

    # filter the nodes
    def check_fn(key, value):
        if has_node(key):
            return value
        else:
            logging.debug('No segments for node %s. Deleting' % key)
            return None
    map_over_dict_with_deletions(nodes, check_fn)

def remove_unconnected_segments(reactions, nodes):
    """Check for segments with no connected nodes.

    Arguments
    ---------

    reactions: A collection (dict) of reactions.

    nodes: A collection (dict) of nodes.

    """
    # use a set to keep track of connected nodes
    node_ids = set(nodes.keys())
    is_node = lambda x: x in node_ids

    # filter the segments
    def check_fn(key, value):
        if is_node(value['from_node_id']) and is_node(value['to_node_id']):
            return value
        else:
            logging.debug('Missing node for segment %s. Deleting' % key)
            return None
    for reaction in reactions.values():
        map_over_dict_with_deletions(reaction['segments'], check_fn)

def remove_reactions_with_missing_metabolites(reactions, nodes):
    """Check for reactions that do not have all of their metabolites.

    Arguments
    ---------

    reactions: A collection (dict) of reactions.

    nodes: A collection (dict) of nodes.

    """
    # filter the reactions
    def check_fn(reaction_id, reaction):
        # get the metabolites
        metabolite_ids = {x['bigg_id'] for x in reaction['metabolites']}
        # look for matching segments
        for segment_id, segment in reaction['segments'].items():
            # find node
            for n in 'from_node_id', 'to_node_id':
                node = nodes[segment[n]]
                try:
                    metabolite_ids.remove(node['bigg_id'])
                except KeyError:
                    pass
        if len(metabolite_ids) > 0:
            logging.info('Deleting reaction %s with missing metabolites %s' % (reaction['bigg_id'],
                                                                               str(list(metabolite_ids))))
            return None
        else:
            return reaction

    map_over_dict_with_deletions(reactions, check_fn)

# ------------------------------------------------------------------------------
# Functions for converting maps
# ------------------------------------------------------------------------------

def old_map_to_new_schema(the_map, map_name=None, map_description=None):
    """Convert any old map to match the latest schema. Returns a new map.

    Arguments
    ---------

    the_map: An Escher map loaded as a Python object (e.g. json.load('my_map.json')).

    map_name: A name for the map. If a name is already present, this name
    overrides it.

    map_description: A description for the map. If a name is already present,
    this name overrides it.

    """

    def add_default_header(body, name, description):
        """Return a map with header and body."""

        def check_for(var, name):
            """Print a warning if var is None."""
            if var is None:
                logging.warn('No {} for map'.format(name))
                return ''
            return var

        new_id = hashlib.md5(json.dumps(body).encode('utf-8')).hexdigest()
        default_header = {
            "schema": "https://escher.github.io/escher/jsonschema/1-0-0#",
            "homepage": "https://escher.github.io",
            "map_name": check_for(name, 'name'),
            "map_id": new_id,
            "map_description": check_for(description, 'description')
        }
        logging.info('Map has the ID {}'.format(new_id))
        return make_map(default_header, body)

    def add_header_if_missing(a_map):
        """Check for new, 2-level maps, and add the header."""

        if has_header_and_body(a_map):
            use_name = (get_header(a_map)['map_name'] if map_name is None else map_name)
            use_description = (get_header(a_map)['map_description']
                               if map_description is None else map_description)
            return add_default_header(get_body(a_map), use_name, use_description)
        elif is_valid_body(a_map):
            return add_default_header(a_map, map_name, map_description)
        else:
            raise Exception('The map provided cannot be converted. It is not a valid Escher map.')

    def fix_body(body):
        """Fill in necessary attributes for the body."""

        def get_default(key, _):
            """Get default value for key in body."""
            if key == 'canvas':
                # default canvas
                return {'x': -1440, 'y': -775, 'width': 4320, 'height': 2325}
            else:
                # other defaults
                return {}

        def fix_canvas(canvas):
            """Return a canvas with correct attributes."""
            canvas_keys = ['x', 'y', 'width', 'height']
            cast = {'x': float, 'y': float, 'width': float, 'height': float}
            dict_with_required_elements(canvas, canvas_keys, cast=cast)

        def fix_text_labels(text_labels):
            """Return a list of text_labels with correct attributes."""
            text_label_keys = ['x', 'y', 'text']
            cast = {'x': float, 'y': float}
            collection_of_dicts_with_required_elements(text_labels,
                                                       text_label_keys,
                                                       cast=cast)

        # fix canvas before body so that a default canvas can be added if the
        # canvas is invalid
        try:
            fix_canvas(body['canvas'])
        except KeyError:
            pass
        except MissingDefaultAttribute:
            del body['canvas']

        # fix body
        core_keys = ['nodes', 'reactions', 'text_labels', 'canvas']
        dict_with_required_elements(body, core_keys, get_default)

        # fix text labels
        fix_text_labels(body['text_labels'])

    def fix_nodes(nodes):
        """Fill in necessary attributes for the nodes. Returns nodes."""

        def get_default_node_attr(name, obj):
            """Return default values when possible. Otherwise, raise MissingDefaultAttribute."""
            if name == 'node_is_primary':
                return False
            elif name == 'name':
                return ''
            elif name == 'label_x' and 'x' in obj:
                return obj['x']
            elif name == 'label_y' and 'y' in obj:
                return obj['y']
            else:
                raise MissingDefaultAttribute(name)

        def fix_a_node(node_id, node):
            """Return an updated node, or None if the node is invalid."""

            if not 'node_type' in node:
                logging.debug('Deleting node %s with no node_type' % node_id)
                return None

            elif node['node_type'] == 'metabolite':
                met_keys = ['node_type', 'x', 'y', 'bigg_id', 'name', 'label_x',
                            'label_y', 'node_is_primary']
                cast = {'x': float, 'y': float, 'label_x': float, 'label_y': float}
                try:
                    dict_with_required_elements(node, met_keys, get_default_node_attr, cast=cast)
                    return node
                except MissingDefaultAttribute as e:
                    logging.debug('Deleting node %s with missing attribute %s' % (node_id, e))
                    return None

            elif node['node_type'] in ['multimarker', 'midmarker']:
                marker_keys = ['node_type', 'x', 'y']
                cast = {'x': float, 'y': float}
                try:
                    dict_with_required_elements(node, marker_keys, get_default_node_attr, cast=cast)
                    return node
                except MissingDefaultAttribute as e:
                    logging.debug('Deleting node %s with missing attribute %s' % (node_id, e))
                    return None

            else:
                logging.debug('Deleting node %s with bad node_type %s' % (node_id, node['node_type']))
                return None

        # run fix functions
        map_over_dict_with_deletions(nodes, fix_a_node)

    def fix_reactions(reactions):
        """Fill in necessary attributes for the reactions.

        Returns reactions.

        """

        def get_default_reaction_attr(name, obj):
            """Return default values when possible. Otherwise, raise MissingDefaultAttribute."""
            if name in ['name', 'gene_reaction_rule']:
                return ''
            elif name == 'reversibility':
                return True
            elif name in ['genes', 'metabolites']:
                return []
            else:
                raise MissingDefaultAttribute(name)

        def fix_a_reaction(reaction_id, reaction):
            """Return an updated reaction, or None if the reaction is invalid."""

            def fix_segments(segments):
                """Fix dictionary of segments with correct attributes."""
                def fix_a_segment(segment_id, segment):
                    segment_keys = ['from_node_id', 'to_node_id', 'b1', 'b2']
                    def get_default_segment_attr(key, _):
                        if key in ['b1', 'b2']:
                            return None
                        else:
                            raise MissingDefaultAttribute(key)
                    try:
                        dict_with_required_elements(segment, segment_keys,
                                                    get_default_segment_attr,
                                                    nullable=['b1', 'b2'])
                    except MissingDefaultAttribute as e:
                        logging.debug('Deleting segment %s with missing attribute %s' % (segment_id, e))
                        return None

                    # check the beziers too
                    required_bezier_keys = ['x', 'y']
                    cast = {'x': float, 'y': float}
                    for b in ['b1', 'b2']:
                        try:
                            dict_with_required_elements(segment[b],
                                                        required_bezier_keys,
                                                        cast=cast)
                        except MissingDefaultAttribute as e:
                            logging.debug('Deleting bezier %s with missing attribute %s in segment %s' % (b, e, segment_id))
                            segment[b] = None

                    return segment

                map_over_dict_with_deletions(segments, fix_a_segment)

            def fix_metabolites(metabolites):
                """Return a list of metabolites with correct attributes."""
                metabolite_keys = ['coefficient', 'bigg_id']
                cast = {'coefficient': float}
                return list_of_dicts_with_required_elements(metabolites,
                                                            metabolite_keys,
                                                            cast=cast)

            def fix_genes(genes):
                """Return a list of genes with correct attributes."""
                gene_keys = ['bigg_id', 'name']
                def get_default_gene_attr(name, _):
                    if name == 'name':
                        return ''
                    else:
                        raise MissingDefaultAttribute(name)
                return list_of_dicts_with_required_elements(genes, gene_keys,
                                                            get_default_gene_attr)

            # fix all the attributes
            reaction_keys = ['name', 'bigg_id','reversibility', 'label_x',
                             'label_y', 'gene_reaction_rule', 'genes',
                             'metabolites', 'segments']
            cast = {'label_x': float, 'label_y': float}
            try:
                dict_with_required_elements(reaction, reaction_keys,
                                            get_default_reaction_attr,
                                            cast=cast)
            except MissingDefaultAttribute as e:
                logging.debug('Deleting reaction %s with missing attribute %s' % (reaction_id, e))
                return None

            # fix segments, metabolites, and genes
            fix_segments(reaction['segments'])
            # must have segments
            if len(reaction['segments']) == 0:
                logging.debug('Deleting reaction %s with no segments' % reaction_id)
                return None
            reaction['metabolites'] = fix_metabolites(reaction['metabolites'])
            reaction['genes'] = fix_genes(reaction['genes'])

            return reaction

        # run the fix functions
        map_over_dict_with_deletions(reactions, fix_a_reaction)


    # make sure there is a body and a head
    the_map = add_header_if_missing(the_map)
    body = get_body(the_map)

    # add missing elements to body
    fix_body(body)

    # fix the nodes
    fix_nodes(get_nodes(body))

    # fix the reactions
    fix_reactions(get_reactions(body))

    # delete any nodes with no segment
    remove_unconnected_nodes(get_nodes(body), get_reactions(body))

    # delete segments with no nodes
    remove_unconnected_segments(get_reactions(body), get_nodes(body))

    # delete reactions with missing metabolite segments
    remove_reactions_with_missing_metabolites(get_reactions(body), get_nodes(body))

    return the_map


def apply_id_mappings(the_map, reaction_id_mapping=None,
                      metabolite_id_mapping=None, gene_id_mapping=None):
    """Convert bigg_ids in the map using the mappings dictionaries.

    Arguments
    ---------

    the_map: The Escher map Python object.

    reaction_id_mapping: A dictionary with keys for existing bigg_ids and value for new bigg_ids.

    metabolite_id_mapping: A dictionary with keys for existing bigg_ids and value for new bigg_ids.

    gene_id_mapping: A dictionary with keys for existing bigg_ids and value for new bigg_ids.

    """

    id_key = 'bigg_id'

    def check(a_dict, mapping):
        """Try to change the value for id_key to a new value defined in mapping."""
        try:
            new_id = mapping[a_dict[id_key]]
        except KeyError:
            pass
        else:
            a_dict[id_key] = new_id
        return a_dict

    def apply_mapping_list(a_list, mapping):
        """Use the mapping on each dict in the list. Returns a new list."""
        return [check(x, mapping) for x in a_list]

    def apply_mapping_dict(collection, mapping):
        """Use the mapping on each dict in the collection (dict). Returns the dictionary."""
        for val in collection.values():
            check(val, mapping)
        return collection

    body = get_body(the_map)
    # reactions
    if reaction_id_mapping is not None:
        apply_mapping_dict(get_reactions(body), reaction_id_mapping)
    # metabolites
    if metabolite_id_mapping is not None:
        apply_mapping_dict(get_nodes(body), metabolite_id_mapping)
    # genes & metabolites in reactions
    for reaction in get_reactions(body).values():
        if gene_id_mapping is not None:
            reaction['genes'] = apply_mapping_list(reaction['genes'], gene_id_mapping)
        if metabolite_id_mapping is not None:
            reaction['metabolites'] = apply_mapping_list(reaction['metabolites'], metabolite_id_mapping)


def apply_cobra_model_to_map(the_map, model):
    """Apply the COBRA model attributes (descriptive names, gene reaction rules,
    reversibilities) to the map.

    Cleans up unconnected segments and nodes after deleting any nodes and
    reactions not found in the cobra model.

    """

    def apply_model_attributes(a_dict, dict_list, attribute_fns, collection_on='bigg_id'):
        """For each attribute_fn, apply the values from the dict_list to given
        dictionary, using the given IDs. Returns the dictionary, or None if a
        matching object was not found in the DictList.

        """
        try:
            on_id = a_dict[collection_on]
        except KeyError:
            return a_dict
        try:
            dl_object = dict_list.get_by_id(on_id)
        except KeyError:
            logging.info('Could not find %s in model. Deleting.' % on_id)
            return None
        else:
            for attribute_fn in attribute_fns:
                attribute_fn(a_dict, dl_object)

        return a_dict

    def apply_model_attributes_dict(collection, dict_list, attribute_fns,
                                    collection_on='bigg_id'):
        """For each attribute_fn, apply the values from the dict_list to the matching
        results in the collection, using the given IDs. Returns the collection."""
        for key, val in list(collection.items()):
            new_val = apply_model_attributes(val, dict_list, attribute_fns,
                                             collection_on)
            if new_val is None:
                del collection[key]

    def apply_model_attributes_list(a_list, dict_list, attribute_fns,
                                    collection_on='bigg_id'):
        """For each attribute_fn, apply the values from the dict_list to the matching
        results in the list of dicts, using the given IDs. Returns a new list."""
        return [y for y in (apply_model_attributes(x, dict_list, attribute_fns, collection_on) for x in a_list)
                if y is not None]

    def get_attr_fn(attribute):
        """Make a default attribute setting function."""
        def new_attr_fn(val, dl_object):
            try:
                val[attribute] = getattr(dl_object, attribute)
            except AttributeError:
                logging.debug('No %s found in DictList %s' % (attribute, on_id))
        return new_attr_fn

    def set_reversibility(reaction, cobra_reaction):
        reaction['reversibility'] = (cobra_reaction.lower_bound < 0 and cobra_reaction.upper_bound > 0)
        # reverse metabolites if reaction runs in reverse
        rev_mult = (-1 if
                    (cobra_reaction.lower_bound < 0 and cobra_reaction.upper_bound <= 0)
                    else 1)
        # use metabolites from reaction
        reaction['metabolites'] = [{'bigg_id': met.id, 'coefficient': coeff * rev_mult}
                                   for met, coeff in
                                   cobra_reaction.metabolites.items()]

    def set_genes(reaction, cobra_reaction):
        reaction['genes'] = [{'bigg_id': x.id, 'name': x.name} for x in cobra_reaction.genes]

    # vars
    body = get_body(the_map)

    # compare reactions to model
    reaction_attributes = [get_attr_fn('name'), get_attr_fn('gene_reaction_rule'),
                           set_reversibility, set_genes]
    apply_model_attributes_dict(get_reactions(body), model.reactions,
                                reaction_attributes)

    # compare metabolites to model
    metabolite_attributes = [get_attr_fn('name')]
    apply_model_attributes_dict(get_nodes(body), model.metabolites,
                                metabolite_attributes)

    # delete any nodes with no segment
    remove_unconnected_nodes(get_nodes(body), get_reactions(body))

    # delete segments with no nodes
    remove_unconnected_segments(get_reactions(body), get_nodes(body))

    # delete reactions with missing metabolite segments
    remove_reactions_with_missing_metabolites(get_reactions(body), get_nodes(body))


def convert(the_map, model, map_name=None, map_description=None,
            reaction_id_mapping=None, metabolite_id_mapping=None,
            gene_id_mapping=None, debug=False):
    """Convert an Escher map to the latest format using the COBRA model to update
    content. Returns a new map.

    Arguments
    ---------

    the_map: An Escher map loaded as a Python object (e.g. json.load('my_map.json')).

    model: A COBRA model.

    map_name: A name for the map. If a name is already present, this name
    overrides it.

    map_description: A description for the map. If a name is already present,
    this name overrides it.

    reaction_id_mapping: A dictionary with existing reaction IDs as keys and the
    new reaction IDs as values.

    metabolite_id_mapping: A dictionary with existing metabolite IDs as keys and the
    new metabolite IDs as values.

    gene_id_mapping: A dictionary with existing gene IDs as keys and the new
    gene IDs as values.

    debug: Check the map against the schema at some intermediate steps.

    """

    # make sure everything is up-to-date
    new_map = old_map_to_new_schema(the_map, map_name=map_name,
                                    map_description=map_description)
    if debug:
        validate_map(new_map)

    # apply the ids mappings
    apply_id_mappings(new_map, reaction_id_mapping, metabolite_id_mapping,
                      gene_id_mapping)
    if debug:
        validate_map(new_map)

    # apply the new model
    apply_cobra_model_to_map(new_map, model)

    validate_map(new_map)
    return new_map


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from escher.version import __version__, __schema_version__, __map_model_version__
import os
import re
from os.path import dirname, realpath, join

root_directory = realpath(join(dirname(__file__), '..'))
top_directory = realpath(join(dirname(__file__), '..', '..'))

_escher_local = {
    'builder_css': 'escher/static/escher/builder.css',
    'builder_css_min': 'escher/static/escher/builder.min.css',
    'escher': 'escher/static/escher/escher.js',
    'escher_min': 'escher/static/escher/escher.min.js',
    'logo': 'escher/static/img/escher-logo@2x.png',
    'favicon': 'escher/static/img/favicon.ico',
    'homepage_js': 'escher/static/homepage/main.js',
    'homepage_css': 'escher/static/homepage/main.css',
    'server_index': '../%s/%s/index.json' % (__schema_version__, __map_model_version__),
    'map_download': '../%s/%s/maps/' % (__schema_version__, __map_model_version__),
    'model_download': '../%s/%s/models/' % (__schema_version__, __map_model_version__),
}

_escher_web = {
    'server_index': '%s/%s/index.json' % (__schema_version__, __map_model_version__),
    'map_download': '%s/%s/maps/' % (__schema_version__, __map_model_version__),
    'model_download': '%s/%s/models/' % (__schema_version__, __map_model_version__),
    'favicon': 'escher/static/img/favicon.ico',
}

_dependencies = {
    'd3': 'escher/static/lib/d3.min.js',
    'boot_js': 'escher/static/lib/bootstrap.min.js',
    'boot_css': 'escher/static/lib/bootstrap.min.css',
    'jquery': 'escher/static/lib/jquery.min.js',
    'boot_font_eot': 'escher/static/fonts/glyphicons-halflings-regular.eot',
    'boot_font_svg': 'escher/static/fonts/glyphicons-halflings-regular.svg',
    'boot_font_woff': 'escher/static/fonts/glyphicons-halflings-regular.woff',
    'boot_font_ttf': 'escher/static/fonts/glyphicons-halflings-regular.ttf',
}

_dependencies_cdn = {
    'builder_css': '//npmcdn.com/escher-vis@%s/css/dist/builder.css' % __version__,
    'builder_css_min': '//npmcdn.com/escher-vis@%s/css/dist/builder.min.css' % __version__,
    'escher': '//npmcdn.com/escher-vis@%s/js/dist/escher.js' % __version__,
    'escher_min': '//npmcdn.com/escher-vis@%s/js/dist/escher.min.js' % __version__,
    'd3': '//cdnjs.cloudflare.com/ajax/libs/d3/3.5.14/d3.min.js',
    'boot_js': '//cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.6/js/bootstrap.min.js',
    'boot_css': '//cdnjs.cloudflare.com/ajax/libs/bootswatch/3.3.6/simplex/bootstrap.min.css',
    'jquery': '//cdnjs.cloudflare.com/ajax/libs/jquery/2.2.0/jquery.min.js',
}

_links = {
    'escher_root': '//escher.github.io/',
    'github': '//github.com/zakandrewking/escher/',
    'github_releases': '//github.com/zakandrewking/escher/releases',
    'documentation': '//escher.readthedocs.org/',
}

# external dependencies
names = list(_escher_local.keys()) + list(_escher_web.keys()) + list(_dependencies.keys()) + list(_links.keys())

def get_url(name, source='web', local_host=None, protocol=None):
    """Get a url.

    Arguments
    ---------

    name: The name of the URL. Options are available in urls.names.

    source: Either 'web' or 'local'. Cannot be 'local' for external links.

    protocol: The protocol can be 'http', 'https', or None which indicates a
    'protocol relative URL', as in //escher.github.io. Ignored if source is
    local.

    local_host: A host url, including the protocol. e.g. http://localhost:7778.

    """
    if source not in ['web', 'local']:
        raise Exception('Bad source: %s' % source)

    if protocol not in [None, 'http', 'https']:
        raise Exception('Bad protocol: %s' % protocol)

    if protocol is None:
        protocol = ''
    else:
        protocol = protocol + ':'

    def apply_local_host(url):
        return '/'.join([local_host.rstrip('/'), url.lstrip('/')])

    # escher
    if name in _escher_local and source == 'local':
        if local_host is not None:
            return apply_local_host(_escher_local[name])
        return _escher_local[name]
    elif name in _escher_web and source == 'web':
        return protocol + '/'.join([_links['escher_root'].rstrip('/'),
                                    _escher_web[name].lstrip('/')])
    # links
    elif name in _links:
        if source=='local':
            raise Exception('Source cannot be "local" for external links')
        return protocol + _links[name]
    # local dependencies
    elif name in _dependencies and source == 'local':
        if local_host is not None:
            return apply_local_host(_dependencies[name])
        return _dependencies[name]
    # cdn dependencies
    elif name in _dependencies_cdn and source == 'web':
        return protocol + _dependencies_cdn[name]

    raise Exception('name not found')



import sys

# user input for python 2 and 3
try:
    import __builtin__
    input = getattr(__builtin__, 'raw_input')
except (ImportError, AttributeError):
    pass

def query_yes_no(question):
    """Ask a yes/no question via input() and return their answer.

    Returns True for yes or False for no.

    Arguments
    ---------

    question: A string that is presented to the user.


    Adapted from http://stackoverflow.com/questions/3041986/python-command-line-yes-no-input.

    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    prompt = " [y/n] "

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        try:
            return valid[choice]
        except KeyError:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")



import json
from os.path import join, dirname

def get_full_version(main_version, post_version=None):
    """Generate a PEP440 compliant version with an optional post-release."""
    if post_version is None:
        return main_version
    else:
        return '%s.post%s' % (main_version, post_version)

with open(join(dirname(__file__), 'package.json'), 'r') as f:
    package = json.load(f)

# software version
__version__ = package['version']
# Escher schema and map version. The schema version determines the version of
# the map specification, and the map version is an additional level so that maps
# can be released with the same spec but new layouts.
__schema_version__ = package['schema_version']
__map_model_version__ = package['map_model_version']

# post-release version (required by PyPI & PEP440)
__post_version__ = None
# full version with post-release
__full_version__ = get_full_version(__version__, __post_version__)



# -*- coding: utf-8 -*-

import sys
from sys import argv
from subprocess import call
import threading
import webbrowser
import os
from shutil import copy, move, rmtree
from os.path import join, dirname, realpath, exists
from glob import glob
import re

try:
    from setuptools import setup, Command
    from setuptools.command.sdist import sdist as SDistCommand
    from setuptools.command.bdist import bdist as BDistCommand
    from setuptools.command.upload import upload as UploadCommand
except ImportError:
    from distutils.core import setup, Command

directory = dirname(realpath(__file__))
sys.path.insert(0, join(directory, 'escher'))
version = __import__('version').__version__
full_version = __import__('version').__full_version__
package = __import__('version').package
port = 8789

class CleanCommand(Command):
    description = "Custom clean command that removes static site"
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        def remove_if(x):
            if exists(x): rmtree(x)
        remove_if(join(directory, 'build'))
        remove_if(join(directory, 'dist'))
        # remove site files
        remove_if(join(directory, '..', 'builder'))
        for f in glob(join(directory, '..', 'index.html')):
            os.remove(f)
        print('done cleaning')


class BuildGHPagesCommand(Command):
    description = "Custom build command that generates static site, and copies escher libs"
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
       pass
    def run(self):
        # generate the static site
        call(['python', join('escher', 'generate_index.py')])
        call(['python', join('escher', 'static_site.py')])
        print('Done building gh-pages')


class TestCommand(Command):
    description = "Custom test command that runs pytest"
    user_options = [('noweb', None, 'Skip run tests that require the Escher website')]
    def initialize_options(self):
        self.noweb = False
    def finalize_options(self):
        pass
    def run(self):
        import pytest
        if self.noweb:
            exit_code = pytest.main(['-m', 'not web'])
        else:
            exit_code = pytest.main([])
        sys.exit(exit_code)


setup(
    name='Escher',
    version=full_version,
    author=package['author'],
    url=package['homepage'],
    description=package['description'],
    keywords=', '.join(package['keywords']),
    license=package['license'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Visualization',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Operating System :: OS Independent'
    ],
    packages=['escher'],
    package_data={'escher': ['package.json', 'static/escher/*',
                             'static/fonts/*', 'static/homepage/*',
                             'static/img/*', 'static/lib/*', 'templates/*']},
    install_requires=['Jinja2>=2.7.3',
                      'tornado>=4.0.2',
                      'pytest>=2.6.2',
                      'jsonschema>=2.4.0'],
    extras_require={'docs': ['sphinx>=1.2',
                             'sphinx-rtd-theme>=0.1.6'],
                    'all': ['sphinx>=1.2',
                            'sphinx-rtd-theme>=0.1.6',
                            'ipython>=4.0.2',
                            'jupyter>=1.0.0',
                            'cobra>=0.3.0',
                            'wheel>=0.24.0',
                            'twine>=1.5.0'] },
    cmdclass={'clean': CleanCommand,
              'build_gh': BuildGHPagesCommand,
              'test': TestCommand}
)



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from escher.quick_server import serve_and_open
from escher.urls import get_url
from escher.appdirs import user_cache_dir
from escher.generate_index import index
from escher.version import __schema_version__, __map_model_version__
from escher.util import query_yes_no
from escher.escape import json_dump_and_escape, escape_json_or_null

import os
from os.path import dirname, basename, abspath, join, isfile, isdir, exists
from warnings import warn
try:
    from urllib.request import urlopen
    from urllib.error import URLError
except:
    from urllib2 import urlopen, URLError
import json
import shutil
import re
from jinja2 import Environment, PackageLoader, Template
import codecs
import random
import string
from tornado.escape import url_escape
import shutil

# set up jinja2 template location
env = Environment(loader=PackageLoader('escher', 'templates'))

# cache management

def get_cache_dir(versioned=True, name=None):
    """Get the cache dir as a string, and make the directory if it does not already
    exist.

    :param Boolean versioned: Whether to return the versioned path in the
                              cache. Escher maps for the latest version of
                              Escher are found in the versioned directory
                              (versioned = True), but maps for previous versions
                              of Escher can be found by visiting the parent
                              directory (versioned = False).

    :param string name: An optional subdirectory within the cache. If versioned
                        is False, then name is ignored.

    """
    cache_dir = user_cache_dir('escher', appauthor='Zachary King')
    # add version
    if versioned:
        cache_dir = join(cache_dir, __schema_version__, __map_model_version__)
        # add subdirectory
        if name is not None:
            cache_dir = join(cache_dir, name)
    try:
        os.makedirs(cache_dir)
    except OSError:
        pass
    return cache_dir

def clear_cache(different_cache_dir=None, ask=True):
    """Empty the contents of the cache directory, including all versions of all maps
    and models.

    :param string different_cache_dir: (Optional) The directory of another
                                       cache. This is mainly for testing.

    :param Boolean ask: Whether to ask before deleting.

    """
    if ask and not query_yes_no('Are you sure you want to delete the contents of the cache?'):
        return

    if different_cache_dir is None:
        cache_dir = get_cache_dir(versioned=False)
    else:
        cache_dir = different_cache_dir

    for root, dirs, files in os.walk(cache_dir):
        for f in files:
            os.unlink(join(root, f))
        for d in dirs:
            shutil.rmtree(join(root, d))

def local_index(cache_dir=get_cache_dir()):
    return index(cache_dir)

def list_cached_maps():
    """Return a list of all cached maps."""
    return local_index()['maps']

def list_cached_models():
    """Return a list of all cached models."""
    return local_index()['models']

# server management

def server_index():
    url = get_url('server_index', source='web', protocol='https')
    try:
        download = urlopen(url)
    except URLError:
        raise URLError('Could not contact Escher server')
    data = _decode_response(download)
    index = json.loads(data)
    return index

def list_available_maps():
    """Return a list of all maps available on the server"""
    return server_index()['maps']

def list_available_models():
    """Return a list of all models available on the server"""
    return server_index()['models']

# download maps and models

def _json_for_name(name, kind, cache_dir):
    # check the name
    name = name.replace('.json', '')

    def match_in_index(name, index, kind):
        return [(obj['organism'], obj[kind + '_name']) for obj in index[kind + 's']
                if obj[kind + '_name'] == name]

    # first check the local index
    match = match_in_index(name, local_index(cache_dir=cache_dir), kind)
    if len(match) == 0:
        path = None
    else:
        org, name = match[0]
        path = join(cache_dir, kind + 's', org, name + '.json')

    if path:
        # load the file
        with open(path, 'rb') as f:
            return f.read().decode('utf-8')
    # if the file is not present attempt to download
    else:
        try:
            index = server_index()
        except URLError:
            raise Exception(('Could not find the %s %s in the cache, and could '
                             'not connect to the Escher server' % (kind, name)))
        match = match_in_index(name, server_index(), kind)
        if len(match) == 0:
            raise Exception('Could not find the %s %s in the cache or on the server' % (kind, name))
        org, name = match[0]
        url = (get_url(kind + '_download', source='web', protocol='https') +
               '/'.join([url_escape(x, plus=False) for x in [org, name + '.json']]))
        warn('%s not in cache. Attempting download from %s' % (kind.title(), url))
        try:
            download = urlopen(url)
        except URLError:
            raise ValueError('No %s found in cache or at %s' % (kind, url))
        data = _decode_response(download)
        # save the file
        org_path = join(cache_dir, kind + 's', org)
        try:
            os.makedirs(org_path)
        except OSError:
            pass
        with open(join(org_path, name + '.json'), 'w') as outfile:
            outfile.write(data)
        return data

def model_json_for_name(model_name, cache_dir=get_cache_dir()):
    return _json_for_name(model_name, 'model', cache_dir)

def map_json_for_name(map_name, cache_dir=get_cache_dir()):
    return _json_for_name(map_name, 'map', cache_dir)

# helper functions

def _get_an_id():
    return (''.join(random.choice(string.ascii_lowercase)
                    for _ in range(10)))

def _decode_response(download):
    """Decode the urllib.response.addinfourl response."""
    data = download.read()
    try:                    # Python 2
        encoding = download.headers.getparam('charset')
    except AttributeError:  # Python 3
        encoding = download.headers.get_param('charset')
    if encoding:
        data = data.decode(encoding)
    else:
        data = data.decode('utf-8')
    return data

def _load_resource(resource, name, safe=False):
    """Load a resource that could be a file, URL, or json string."""
    # if it's a url, download it
    if resource.startswith('http://') or resource.startswith('https://'):
        try:
            download = urlopen(resource)
        except URLError as err:
            raise err
        else:
            return _decode_response(download)
    # if it's a filepath, load it
    if exists(resource):
        if (safe):
            raise Exception('Cannot load resource from file with safe mode enabled.')
        try:
            with open(resource, 'rb') as f:
                loaded_resource = f.read().decode('utf-8')
            _ = json.loads(loaded_resource)
        except ValueError as err:
            raise ValueError('%s not a valid json file' % name)
        else:
            return loaded_resource
    # try to validate the json
    try:
        _ = json.loads(resource)
    except ValueError as err:
        raise ValueError('Could not load %s. Not valid json, url, or filepath' % name)
    else:
        return resource
    raise Exception('Could not load %s.' % name)


class Builder(object):
    """A metabolic map that can be viewed, edited, and used to visualize data.

    This map will also show metabolic fluxes passed in during consruction.  It
    can be viewed as a standalone html inside a browswer. Alternately, the
    respresentation inside an IPython notebook will also display the map.

    Maps are stored in json files and are stored in a cache directory. Maps
    which are not found will be downloaded from a map repository if found.

    :param map_name:

        A string specifying a map to be downloaded from the Escher web server,
        or loaded from the cache.

    :param map_json:

        A JSON string, or a file path to a JSON file, or a URL specifying a JSON
        file to be downloaded.

    :param model: A Cobra model.

    :param model_name:

        A string specifying a model to be downloaded from the Escher web server,
        or loaded from the cache.

    :param model_json:

        A JSON string, or a file path to a JSON file, or a URL specifying a JSON
        file to be downloaded.

    :param embedded_css:

        The CSS (as a string) to be embedded with the Escher SVG.

    :param reaction_data:

        A dictionary with keys that correspond to reaction ids and values that
        will be mapped to reaction arrows and labels.

    :param metabolite_data:

        A dictionary with keys that correspond to metabolite ids and values that
        will be mapped to metabolite nodes and labels.

    :param gene_data:

        A dictionary with keys that correspond to gene ids and values that will
        be mapped to corresponding reactions.

    :param local_host:

        A hostname that will be used for any local files. This is generally used
        for using the notebook offline and for testing in the IPython Notebook
        with modified Escher code. An example value for local_host is
        'http://localhost:7778/'.

    :param id:

        Specify an id to make the javascript data definitions unique. A random
        id is chosen by default.

    :param safe:

        If True, then loading files from the filesytem is not allowed. This is
        to ensure the safety of using Builder within a web server.

    **Keyword Arguments**

    These are defined in the Javascript API:

        - use_3d_transform
        - enable_search
        - fill_screen
        - zoom_to_element
        - full_screen_button
        - starting_reaction
        - unique_map_id
        - primary_metabolite_radius
        - secondary_metabolite_radius
        - marker_radius
        - gene_font_size
        - hide_secondary_metabolites
        - show_gene_reaction_rules
        - hide_all_labels
        - reaction_styles
        - reaction_compare_style
        - reaction_scale
        - reaction_no_data_color
        - reaction_no_data_size
        - and_method_in_gene_reaction_rule
        - metabolite_styles
        - metabolite_compare_style
        - metabolite_scale
        - metabolite_no_data_color
        - metabolite_no_data_size
        - identifiers_on_map
        - highlight_missing
        - allow_building_duplicate_reactions
        - cofactors

    All keyword arguments can also be set on an existing Builder object
    using setter functions, e.g.:

    .. code:: python

        my_builder.set_reaction_styles(new_styles)

    """

    def __init__(self, map_name=None, map_json=None, model=None,
        model_name=None, model_json=None, embedded_css=None,
        reaction_data=None, metabolite_data=None, gene_data=None,
        local_host=None, id=None, safe=False, **kwargs):

        self.safe = safe

        # load the map
        self.map_name = map_name
        self.map_json = map_json
        self.loaded_map_json = None
        if map_name and map_json:
            warn('map_json overrides map_name')
        self._load_map()
        # load the model
        self.model = model
        self.model_name = model_name
        self.model_json = model_json
        self.loaded_model_json = None
        if sum([x is not None for x in (model, model_name, model_json)]) >= 2:
            warn('model overrides model_json, and both override model_name')
        self._load_model()
        # set the args
        self.reaction_data = reaction_data
        self.metabolite_data = metabolite_data
        self.gene_data = gene_data
        self.local_host = local_host

        # remove illegal characters from css
        try:
            self.embedded_css = (embedded_css.replace('\n', ''))
        except AttributeError:
            self.embedded_css = None
        # make the unique id
        self.the_id = _get_an_id() if id is None else id

        # set up the options
        self.options = ['use_3d_transform',
                        'enable_search',
                        'fill_screen',
                        'zoom_to_element',
                        'full_screen_button',
                        'starting_reaction',
                        'unique_map_id',
                        'primary_metabolite_radius',
                        'secondary_metabolite_radius',
                        'marker_radius',
                        'gene_font_size',
                        'hide_secondary_metabolites',
                        'show_gene_reaction_rules',
                        'hide_all_labels',
                        'reaction_styles',
                        'reaction_compare_style',
                        'reaction_scale',
                        'reaction_no_data_color',
                        'reaction_no_data_size',
                        'and_method_in_gene_reaction_rule',
                        'metabolite_styles',
                        'metabolite_compare_style',
                        'metabolite_scale',
                        'metabolite_no_data_color',
                        'metabolite_no_data_size',
                        'identifiers_on_map',
                        'highlight_missing',
                        'allow_building_duplicate_reactions',
                        'cofactors']
        def get_getter_setter(o):
            """Use a closure."""
            # create local fget and fset functions
            fget = lambda self: getattr(self, '_%s' % o)
            fset = lambda self, value: setattr(self, '_%s' % o, value)
            return fget, fset
        for option in self.options:
            fget, fset = get_getter_setter(option)
            # make the setter
            setattr(self.__class__, 'set_%s' % option, fset)
            # add property to self
            setattr(self.__class__, option, property(fget))
            # add corresponding local variable
            setattr(self, '_%s' % option, None)

        # set the kwargs
        for key, val in kwargs.items():
            try:
                getattr(self, 'set_%s' % key)(val)
            except AttributeError:
                print('Unrecognized keywork argument %s' % key)


    def _load_model(self):
        """Load the model.

        Try first from self.model, second from self.model_json, and
        third from from self.model_name.

        """
        if self.model is not None:
            try:
                import cobra.io
            except ImportError:
                raise Exception(('The COBRApy package must be available to load '
                                 'a COBRA model object'))
            self.loaded_model_json = cobra.io.to_json(self.model)
        elif self.model_json is not None:
            self.loaded_model_json = _load_resource(self.model_json,
                                                    'model_json',
                                                    safe=self.safe)
        elif self.model_name is not None:
            self.loaded_model_json = model_json_for_name(self.model_name)


    def _load_map(self):
        """Load the map from input map_json using _load_resource, or, secondarily,
           from map_name.

        """
        if self.map_json is not None:
            self.loaded_map_json = _load_resource(self.map_json,
                                                  'map_json',
                                                  safe=self.safe)
        elif self.map_name is not None:
            self.loaded_map_json = map_json_for_name(self.map_name)


    def _get_html(self, js_source='web', menu='none', scroll_behavior='pan',
                  html_wrapper=False, enable_editing=False, enable_keys=False,
                  minified_js=True, fill_screen=False, height='800px',
                  never_ask_before_quit=False, static_site_index_json=None,
                  protocol=None):
        """Generate the Escher HTML.

        Arguments
        --------

        js_source: Can be one of the following:
            'web' - (Default) use js files from escher.github.io.
            'local' - use compiled js files in the local escher installation. Works offline.
            'dev' - use the local, uncompiled development files. Works offline.

        menu: Menu bar options include:
            'none' - (Default) No menu or buttons.
            'zoom' - Just zoom buttons (does not require bootstrap).
            'all' - Menu and button bar (requires bootstrap).

        scroll_behavior: Scroll behavior options:
            'pan' - (Default) Pan the map.
            'zoom' - Zoom the map.
            'none' - No scroll events.

        minified_js: If True, use the minified version of JavaScript and CSS
        files.

        html_wrapper: If True, return a standalone html file.

        enable_editing: Enable the editing modes (build, rotate, etc.).

        enable_keys: Enable keyboard shortcuts.

        height: The height of the HTML container.

        never_ask_before_quit: Never display an alert asking if you want to
        leave the page. By default, this message is displayed if enable_editing
        is True.

        static_site_index_json: The index, as a JSON string, for the static
        site. Use javascript to parse the URL options. Used for
        generating static pages (see static_site.py).

        protocol: The protocol can be 'http', 'https', or None which indicates a
        'protocol relative URL', as in //escher.github.io. Ignored if source is
        local.

        """

        if js_source not in ['web', 'local', 'dev']:
            raise Exception('Bad value for js_source: %s' % js_source)

        if menu not in ['none', 'zoom', 'all']:
            raise Exception('Bad value for menu: %s' % menu)

        if scroll_behavior not in ['pan', 'zoom', 'none']:
            raise Exception('Bad value for scroll_behavior: %s' % scroll_behavior)

        content = env.get_template('content.html')

        # if height is not a string
        if type(height) is int:
            height = "%dpx" % height
        elif type(height) is float:
            height = "%fpx" % height
        elif type(height) is str:
            height = str(height)

        # set the proper urls
        url_source = 'local' if (js_source=='local' or js_source=='dev') else 'web'
        local_host = self.local_host

        # get the urls
        d3_url = get_url('d3', url_source, local_host, protocol)
        escher_url = get_url(('escher_min' if minified_js else 'escher'),
                             url_source, local_host, protocol)
        if menu == 'all':
            jquery_url = get_url('jquery', url_source, local_host, protocol)
            boot_css_url = get_url('boot_css', url_source, local_host, protocol)
            boot_js_url = get_url('boot_js', url_source, local_host, protocol)
        else:
            jquery_url = boot_css_url = boot_js_url = None
        escher_css_url = get_url(('builder_css_min' if minified_js else 'builder_css'),
                                 url_source, local_host, protocol)
        favicon_url = get_url('favicon', url_source, local_host, protocol)
        # for static site
        map_download_url = get_url('map_download', url_source, local_host, protocol)
        model_download_url = get_url('model_download', url_source, local_host, protocol)

        # local host
        lh_string = ('' if local_host is None else
                     local_host.rstrip('/') + '/')

        # options
        options = {
            'menu': menu,
            'enable_keys': enable_keys,
            'enable_editing': enable_editing,
            'scroll_behavior': scroll_behavior,
            'fill_screen': fill_screen,
            'never_ask_before_quit': never_ask_before_quit,
            'reaction_data': self.reaction_data,
            'metabolite_data': self.metabolite_data,
            'gene_data': self.gene_data,
        }
        # Add the specified options
        for option in self.options:
            val = getattr(self, option)
            if val is None:
                continue
            options[option] = val

        html = content.render(
            # standalone
            title='Escher ' + ('Builder' if enable_editing else 'Viewer'),
            jquery_url=jquery_url,
            boot_js_url=boot_js_url,
            boot_css_url=boot_css_url,
            escher_css_url=escher_css_url,
            favicon_url=favicon_url,
            # content
            wrapper=html_wrapper,
            height=height,
            id=self.the_id,
            d3_url=d3_url,
            escher_url=escher_url,
            # dump json
            id_json=json_dump_and_escape(self.the_id),
            options_json=json_dump_and_escape(options),
            map_download_url_json=json_dump_and_escape(map_download_url),
            model_download_url_json=json_dump_and_escape(model_download_url),
            builder_embed_css_json=json_dump_and_escape(self.embedded_css),
            # alreay json
            map_data_json=escape_json_or_null(self.loaded_map_json),
            model_data_json=escape_json_or_null(self.loaded_model_json),
            static_site_index_json=escape_json_or_null(static_site_index_json),
        )

        return html


    def display_in_notebook(self, js_source='web', menu='zoom', scroll_behavior='none',
                            minified_js=True, height=500):
        """Embed the Map within the current IPython Notebook.

        :param string js_source:

            Can be one of the following:

            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher installation. Works offline.
            - *dev* - No longer necessary with source maps. This now gives the
                      same behavior as 'local'.

        :param string menu: Menu bar options include:

            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - Note: The *all* menu option does not work in an IPython notebook.

        :param string scroll_behavior: Scroll behavior options:

            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* - (Default) No scroll events.

        :param Boolean minified_js:

            If True, use the minified version of JavaScript and CSS files.

        :param height: Height of the HTML container.

        """
        html = self._get_html(js_source=js_source, menu=menu, scroll_behavior=scroll_behavior,
                              html_wrapper=False, enable_editing=False, enable_keys=False,
                              minified_js=minified_js, fill_screen=False, height=height,
                              never_ask_before_quit=True)
        if menu=='all':
            raise Exception("The 'all' menu option cannot be used in an IPython notebook.")
        # import here, in case users don't have requirements installed
        try:
            from IPython.display import HTML
        except ImportError:
            raise Exception('You need to be using the IPython notebook for this function to work')
        return HTML(html)


    def display_in_browser(self, ip='127.0.0.1', port=7655, n_retries=50, js_source='web',
                           menu='all', scroll_behavior='pan', enable_editing=True, enable_keys=True,
                           minified_js=True, never_ask_before_quit=False):
        """Launch a web browser to view the map.

        :param ip: The IP address to serve the map on.

        :param port:

            The port to serve the map on. If specified the port is occupied,
            then a random free port will be used.

        :param int n_retries:

            The number of times the server will try to find a port before
            quitting.

        :param string js_source:

            Can be one of the following:

            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher installation. Works offline.
            - *dev* - No longer necessary with source maps. This now gives the
                      same behavior as 'local'.

        :param string menu: Menu bar options include:

            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - *all* (Default) - Menu and button bar (requires Bootstrap).

        :param string scroll_behavior: Scroll behavior options:

            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* (Default) - No scroll events.

        :param Boolean enable_editing: Enable the map editing modes.

        :param Boolean enable_keys: Enable keyboard shortcuts.

        :param Boolean minified_js:

            If True, use the minified version of JavaScript and CSS files.

        :param Boolean never_ask_before_quit:

            Never display an alert asking if you want to leave the page. By
            default, this message is displayed if enable_editing is True.

        """
        html = self._get_html(js_source=js_source, menu=menu, scroll_behavior=scroll_behavior,
                              html_wrapper=True, enable_editing=enable_editing, enable_keys=enable_keys,
                              minified_js=minified_js, fill_screen=True, height="100%",
                              never_ask_before_quit=never_ask_before_quit)
        serve_and_open(html, ip=ip, port=port, n_retries=n_retries)


    def save_html(self, filepath=None, overwrite=False, js_source='web',
                  protocol='https', menu='all', scroll_behavior='pan',
                  enable_editing=True, enable_keys=True, minified_js=True,
                  never_ask_before_quit=False, static_site_index_json=None):
        """Save an HTML file containing the map.

        :param string filepath:

            The HTML file will be saved to this location. When js_source is
            'local', then a new directory will be created with this name.

        :param Boolean overwrite:

            Overwrite existing files.

        :param string js_source:

            Can be one of the following:

            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher
                        installation. Works offline. To make the dependencies
                        available to the downloaded file, a new directory will
                        be made with the name specified by filepath.
            - *dev* - No longer necessary with source maps. This now gives the
                      same behavior as 'local'.

        :param string protocol:

            The protocol can be 'http', 'https', or None which indicates a
            'protocol relative URL', as in //escher.github.io. Ignored if source
            is local.

        :param string menu: Menu bar options include:

            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - *all* (Default) - Menu and button bar (requires Bootstrap).

        :param string scroll_behavior: Scroll behavior options:

            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* (Default) - No scroll events.

        :param Boolean enable_editing: Enable the map editing modes.

        :param Boolean enable_keys: Enable keyboard shortcuts.

        :param Boolean minified_js:

            If True, use the minified version of JavaScript and CSS files.

        :param number height: Height of the HTML container.

        :param Boolean never_ask_before_quit:

            Never display an alert asking if you want to leave the page. By
            default, this message is displayed if enable_editing is True.

        :param string static_site_index_json:

            The index, as a JSON string, for the static site. Use javascript
            to parse the URL options. Used for generating static pages (see
            static_site.py).

        """

        if js_source in ['local', 'dev']:
            if filepath is None:
                raise Exception('Must provide a filepath when js_source is not "web"')

            # make a directory
            directory = re.sub(r'\.html$', '', filepath)
            if exists(directory):
                if not overwrite:
                    raise Exception('Directory already exists: %s' % directory)
            else:
                os.makedirs(directory)
            # add dependencies to the directory
            escher = get_url('escher_min' if minified_js else 'escher', 'local')
            builder_css = get_url('builder_css_min' if minified_js else 'builder_css', 'local')
            d3 = get_url('d3', 'local')
            favicon = get_url('favicon', 'local')
            if menu == 'all':
                boot_css = get_url('boot_css', 'local')
                boot_js = get_url('boot_js', 'local')
                jquery = get_url('jquery', 'local')
                boot_font_eot = get_url('boot_font_eot', 'local')
                boot_font_svg = get_url('boot_font_svg', 'local')
                boot_font_woff = get_url('boot_font_woff', 'local')
                boot_font_ttf = get_url('boot_font_ttf', 'local')
            else:
                boot_css = boot_js = jquery = None
                boot_font_eot = boot_font_svg = boot_font_woff = boot_font_ttf = None
            for path in [escher, builder_css, boot_css, boot_js, jquery, d3,
                         favicon, boot_font_eot, boot_font_svg, boot_font_woff,
                         boot_font_ttf]:
                dest = join(directory, path)
                dest_dir = dirname(dest)
                if not exists(dest_dir):
                    os.makedirs(dest_dir)
                shutil.copy(path, dest)
            filepath = join(directory, 'index.html')
        else:
            if not filepath.endswith('.html'):
                filepath += '.html'
            if exists(filepath) and not overwrite:
                raise Exception('File already exists: %s' % filepath)

        html = self._get_html(js_source=js_source, menu=menu,
                              scroll_behavior=scroll_behavior,
                              html_wrapper=True, enable_editing=enable_editing,
                              enable_keys=enable_keys, minified_js=minified_js,
                              fill_screen=True, height="100%",
                              never_ask_before_quit=never_ask_before_quit,
                              static_site_index_json=static_site_index_json,
                              protocol=protocol)
        if filepath is not None:
            with open(filepath, 'wb') as f:
                f.write(html.encode('utf-8'))
            return filepath
        else:
            from tempfile import mkstemp
            from os import write, close
            os_file, filename = mkstemp(suffix=".html", text=False) # binary
            write(os_file, html.encode('utf-8'))
            close(os_file)
            return filename
